from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

from liblaf.grapes._config import config
from liblaf.grapes.rich.logging import RichHandler

from .filters import FilterLike, as_filter
from .handlers import RichFileHandler
from .helpers import (
    HandlerRestrictedLogger,
    clear_children_stream_handlers,
    init_levels,
    install_excepthook,
    install_unraisablehook,
)

if TYPE_CHECKING:
    from logging import _FilterType


def init(
    *,
    file: str | os.PathLike[str] | None = None,
    filter: FilterLike | None = None,  # noqa: A002
    force: bool = False,
) -> None:
    if file is None:
        file = config.logging.file.get()
    init_levels()
    filter_: _FilterType = as_filter(filter)
    handlers: list[logging.Handler] = [RichHandler()]
    if file is not None:
        handlers.append(RichFileHandler(file))
    for handler in handlers:
        handler.addFilter(filter_)
    clear_children_stream_handlers()
    install_excepthook()
    install_unraisablehook()
    logging.basicConfig(level=logging.NOTSET, handlers=handlers, force=force)
    logging.setLoggerClass(HandlerRestrictedLogger)



from ._array import pdoc_array
from ._chain import chain_custom
from ._dispatch import PdocCustomDispatcher, pdoc_custom
from ._fieldz import pdoc_fieldz
from ._rich_repr import pdoc_rich_repr

__all__ = [
    "PdocCustomDispatcher",
    "chain_custom",
    "pdoc_array",
    "pdoc_custom",
    "pdoc_fieldz",
    "pdoc_rich_repr",
]



import functools
from typing import Unpack

import toolz

from liblaf.grapes._config import config

from ._typing import CustomCallable, WadlerLindigOptions
from .custom import chain_custom, pdoc_custom


def make_kwargs(**kwargs: Unpack[WadlerLindigOptions]) -> WadlerLindigOptions:
    kwargs = toolz.merge(config.pretty.get(), kwargs)
    pdoc_custom_partial: CustomCallable = functools.partial(pdoc_custom, **kwargs)
    kwargs["custom"] = chain_custom(kwargs.get("custom"), pdoc_custom_partial)
    return kwargs



from pathlib import Path

from liblaf.grapes.conf import BaseConfig, Field, field


class ConfigLogging(BaseConfig):
    datefmt: Field[str] = field(default="YYYY-MM-DD HH:mm:ss.SSS", env="LOG_DATEFMT")
    file: Field[Path | None] = field(default=None, env="LOG_FILE")
    hide_frame: Field[list[str]] = field(factory=lambda: ["rich.progress"])
    level: Field[int | str] = field(default="TRACE")
    time_relative: Field[bool] = field(default=True, env="LOG_TIME_RELATIVE")



from liblaf.grapes.conf import BaseConfig

from ._joblib import ConfigJoblib
from ._logging import ConfigLogging
from ._pretty import ConfigPretty
from ._traceback import ConfigTraceback


class Config(BaseConfig):
    joblib: ConfigJoblib
    logging: ConfigLogging
    pretty: ConfigPretty
    traceback: ConfigTraceback


config: Config = Config()



import contextlib
from collections.abc import Generator, Mapping
from typing import Any, Self

import attrs
import cytoolz as toolz

from ._field import Field


class BaseConfigMeta(type):
    def __new__[T: type](
        mcs: type[T],
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        /,
        **kwargs: Any,
    ) -> T:
        cls: T = super().__new__(mcs, name, bases, namespace)
        if "__attrs_attrs__" in namespace:
            return cls
        kwargs.setdefault("frozen", True)
        kwargs.setdefault("init", False)
        cls = attrs.define(cls, **kwargs)
        return cls


class BaseConfig(metaclass=BaseConfigMeta):
    def __init__(self, name: str = "") -> None:
        kwargs: dict[str, Any] = {}
        cls: type[BaseConfig] = type(self)
        cls = attrs.resolve_types(cls)
        for f in attrs.fields(type(self)):
            f: attrs.Attribute
            if not isinstance(f.type, type):
                continue
            name: str = f"{name}.{f.name}" if name else f.name
            if issubclass(f.type, BaseConfig):
                kwargs[f.name] = f.type(name)
            elif issubclass(_unwrap(f.type), Field):
                kwargs[f.name] = f.type(name, **f.metadata)
        self.__attrs_init__(**kwargs)  # pyright: ignore[reportAttributeAccessIssue]

    def get(self) -> dict[str, Any]:
        result: dict[str, Any] = {}
        for f in attrs.fields(type(self)):
            f: attrs.Attribute
            value: Any = getattr(self, f.name)
            if isinstance(value, (BaseConfig, Field)):
                result[f.name] = value.get()
        return result

    def set(self, changes: Mapping[str, Any] = {}, /, **kwargs: Any) -> None:
        changes = toolz.merge(changes, kwargs)
        for key, value in changes.items():
            field: BaseConfig | Field = getattr(self, key)
            field.set(value)

    @contextlib.contextmanager
    def overrides(
        self, changes: Mapping[str, Any] = {}, /, **kwargs: Any
    ) -> Generator[Self]:
        changes = toolz.merge(changes, kwargs)
        with contextlib.ExitStack() as stack:
            for key, value in changes.items():
                field: BaseConfig | Field = getattr(self, key)
                stack.enter_context(field.overrides(value))
            yield self


def _unwrap(obj: Any) -> Any:
    return getattr(obj, "__wrapped__", obj)



import inspect
import logging
import types
from typing import Any

from liblaf.grapes import magic

_PROXY_METHODS: set[str] = {
    "debug",
    "info",
    "warning",
    "error",
    "critical",
    "log",
    "exception",
    "findCaller",
}


class DepthLogger:
    def __getattr__(self, name: str) -> Any:
        if name in _PROXY_METHODS:

            def method(*args, **kwargs) -> None:
                from liblaf.grapes.logging import init

                init()
                depth: int = kwargs.get("stacklevel", 1)
                frame: types.FrameType | None = inspect.currentframe()
                stacklevel: int = 1
                while frame is not None and depth > 0:
                    frame = frame.f_back
                    depth -= 1
                    stacklevel += 1
                    while frame is not None and magic.hidden_from_logging(frame):
                        frame = frame.f_back
                        stacklevel += 1
                logger_name: str | None = None
                if frame is not None:
                    logger_name = frame.f_globals.get("__name__")
                logger: logging.Logger = logging.getLogger(logger_name)
                kwargs["stacklevel"] = stacklevel
                return getattr(logger, name)(*args, **kwargs)

            return method
        frame: types.FrameType | None = inspect.currentframe()
        if frame is not None:
            frame = frame.f_back
        logger_name: str | None = None
        if frame is not None:
            logger_name = frame.f_globals.get("__name__")
        logger: logging.Logger = logging.getLogger(logger_name)
        return getattr(logger, name)


depth_logger = DepthLogger()



from . import filters, handlers, helpers
from ._init import init
from .filters import CompositeFilter, as_filter
from .handlers import RichFileHandler
from .helpers import (
    HandlerRestrictedLogger,
    clear_children_stream_handlers,
    depth_logger,
    init_levels,
    install_excepthook,
    install_unraisablehook,
)

__all__ = [
    "CompositeFilter",
    "HandlerRestrictedLogger",
    "RichFileHandler",
    "as_filter",
    "clear_children_stream_handlers",
    "depth_logger",
    "filters",
    "handlers",
    "helpers",
    "init",
    "init_levels",
    "install_excepthook",
    "install_unraisablehook",
]



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from ._composite import CompositeFilter

if TYPE_CHECKING:
    from logging import _FilterType

    type FilterLike = _FilterType | Mapping[str, int | str]
else:
    type FilterLike = Any


def as_filter(f: FilterLike | None = None, /) -> _FilterType:
    if f is None:
        return CompositeFilter()
    if isinstance(f, Mapping):
        return CompositeFilter(f)
    return f



import logging
import types
from collections.abc import Generator, Iterable

from rich.console import Console, RenderableType
from rich.highlighter import Highlighter, ReprHighlighter
from rich.text import Text

from liblaf.grapes import pretty
from liblaf.grapes.rich._get_console import get_console
from liblaf.grapes.rich.traceback import RichExceptionSummary

from .columns import (
    RichHandlerColumn,
    RichHandlerColumnLevel,
    RichHandlerColumnLocation,
    RichHandlerColumnTime,
)


def _default_columns() -> list[RichHandlerColumn]:
    return [
        RichHandlerColumnTime(),
        RichHandlerColumnLevel(),
        RichHandlerColumnLocation(),
    ]


class RichHandler(logging.Handler):
    columns: list[RichHandlerColumn]
    console: Console
    highlighter: Highlighter

    def __init__(
        self,
        console: Console | None = None,
        *,
        columns: Iterable[RichHandlerColumn] | None = None,
        level: int = logging.NOTSET,
    ) -> None:
        super().__init__(level=level)
        columns = _default_columns() if columns is None else list(columns)
        if console is None:
            console = get_console(stderr=True)
        self.columns = columns
        self.console = console
        self.highlighter = ReprHighlighter()

    def emit(self, record: logging.LogRecord) -> None:
        self.console.print(
            *self._render(record),
            sep="",
            end="",
            overflow="ignore",
            no_wrap=True,
            highlight=False,
            crop=False,
            soft_wrap=False,
        )
        if (exception := self._render_exception(record)) is not None:
            self.console.print(exception)

    def _render(self, record: logging.LogRecord) -> Generator[RenderableType]:
        columns: list[Text] = [column.render(record) for column in self.columns]
        meta: Text = Text(" ").join(columns)
        message: Text = self._render_message(record)
        for line in message.split() or [""]:
            yield meta
            if len(line) > 0:
                yield " "
                yield line
            yield "\n"

    def _render_exception(
        self, record: logging.LogRecord
    ) -> RichExceptionSummary | None:
        if record.exc_info is None:
            return None
        exc_type: type[BaseException] | None
        exc_value: BaseException | None
        traceback: types.TracebackType | None
        exc_type, exc_value, traceback = record.exc_info
        if exc_type is None or exc_value is None:
            return None
        return RichExceptionSummary(exc_type, exc_value, traceback)

    def _render_message(self, record: logging.LogRecord) -> Text:
        if markup := getattr(record, "markup", None):
            return Text.from_markup(markup, style="log.message")
        message: str = record.getMessage()
        if pretty.has_ansi(message):
            return Text.from_ansi(message, style="log.message")
        text: Text = Text(message, style="log.message")
        return self.highlighter(text)



import logging
from collections.abc import Mapping

import toolz


def as_levelno_dict(levels: Mapping[str, int | str]) -> dict[str, int]:
    level_names_mapping: dict[str, int] = logging.getLevelNamesMapping()
    return toolz.valmap(
        lambda level: level if isinstance(level, int) else level_names_mapping[level],
        levels,
    )


def as_levelno(level: int | str) -> int:
    if isinstance(level, int):
        return level
    level_names_mapping: dict[str, int] = logging.getLevelNamesMapping()
    return level_names_mapping[level]



import fieldz
from rich.repr import RichReprResult

from liblaf.grapes.sentinel import MISSING


def rich_repr_fieldz(obj: object) -> RichReprResult:
    for field in fieldz.fields(obj):
        if not field.repr:
            continue
        yield field.name, getattr(obj, field.name, MISSING), field.default



import logging


def clear_children_stream_handlers(logger: str | logging.Logger | None = None) -> None:
    if logger is None or isinstance(logger, str):
        logger = logging.getLogger(logger)
    for child in logger.getChildren():
        for handler in child.handlers[:]:
            if isinstance(handler, logging.StreamHandler):
                child.removeHandler(handler)
        child.propagate = True
        clear_children_stream_handlers(child)
